/*-----------------------------------------------------------------------------
  (C) Zack T Smith

  Object-Oriented C is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published
  by the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
 
  Object-Oriented C is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.
 
  You should have received a copy of the GNU Lesser General Public License
  along with this software.  If not, see <http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------------*/

#ifndef _OOC_COLORS_H
#define _OOC_COLORS_H

#define RGB_TEST_PATTERN (0xdeadf00d)

#define RGB_CLEAR (0xff000000)
#define RGB_BLACK (0)
#define RGB_BLUE (0xff)
#define RGB_BABYBLUE (0x89cff0)
#define RGB_LIGHTBLUE (0x87ceeb)
#define RGB_BRASS (0xc3a368)
#define RGB_BROWN (0x8b4513)
#define RGB_LIGHTBROWN (0xa3866a)
#define RGB_BROWN1 (0xff4040)
#define RGB_BROWN2 (0xf03535)
#define RGB_BROWN3 (0xcd3333)
#define RGB_BROWN4 (0x8b2323)
#define RGB_DARKBROWN (0x654321)
#define RGB_STEELBLUE (0x4682b4)
#define RGB_CADETBLUE (0x5f9ea0)
#define RGB_CHARTREUSE (0x7fff00)
#define RGB_FUSCHIA (0xff00ff)
#define RGB_CORAL (0xff7f50)
#define RGB_CYAN (0xffff)
#define RGB_DARKCYAN (0x8080)
#define RGB_YELLOWGREEN (0x9acd32)
#define RGB_GREENYELLOW (0xadff2f)
#define RGB_MEDIUMGREEN (0xc000)
#define RGB_DARKGREEN (0x84000)
#define RGB_DARKKHAKI (0xbdb76b)
#define RGB_DARKOLIVEGREEN (0x556b2f)
#define RGB_LIGHTORANGE (0xfcd299)
#define RGB_DARKORANGE (0xff8c00)
#define RGB_ORANGE (0xffa500)
#define RGB_ORANGERED (0xff4500)
#define RGB_DODGERBLUE (0x1e90ff)
#define RGB_GOLDENROD (0xdaa520)
#define RGB_GOLD (0xd4af36)
#define RGB_LIGHTGRAY (0xe0e0e0)
#define RGB_GRAY (0xc0c0c0)
#define RGB_DARKGRAY (0x808080)
#define RGB_LIGHTGREEN (0x90ee90)
#define RGB_GREEN (0xff00)
#define RGB_KHAKI (0xf0e68c)
#define RGB_LEMONYELLOW (0xfde910)
#define RGB_DARKYELLOW (0xf6be00)
#define RGB_MAGENTA (0xff00ff)
#define RGB_MAROON (0x800000)
#define RGB_DARKBLUE (0x8b)
#define RGB_NAVYBLUE (0x80)
#define RGB_PINK (0xf77fbe)
#define RGB_LIGHTPURPLE (0xc5b4e3)
#define RGB_PURPLE (0xa020f0)
#define RGB_DARKPURPLE (0x871f78)
#define RGB_LIGHTRED (0xff7276)
#define RGB_RED (0xff0000)
#define RGB_BRICKRED (0xcb4154)
#define RGB_CRIMSON (0xdc143c)
#define RGB_DARKROYALBLUE (0x002366)
#define RGB_ROYALBLUE (0x4169e1)
#define RGB_SALMON (0xfa8072)
#define RGB_TURQUOISE (0x40e0d0)
#define RGB_VIOLET (0xee82ee)
#define RGB_WHITE (0xffffff)
#define RGB_YELLOW (0xffff00)
#define RGB_AMBER (0xffbf00)
#define RGB_TEAL (0x8080)
#define RGB_TEALBLUE (0x007c80)
#define RGB_LIGHTTEAL (0x1ea0a0)
#define RGB_DEEPTEAL (0x033e3e)
#define RGB_DARKTEAL (0x045d5d)

#define RGB_GRAY6 (0x606060)
#define RGB_GRAY8 (0x808080)
#define RGB_GRAY10 (0xa0a0a0)
#define RGB_GRAY12 (0xc0c0c0)
#define RGB_GRAY14 (0xe0e0e0)

#define kColorClear (0xff << 24)
#define kColorBlack (0)
#define kColorBlue (0xff)
#define kColorBrass (0xc3a368)
#define kColorBrown (0x8b4513)
#define kColorCadetBlue (0x5f9ea0)
#define kColorChartreuse (0x7fff00)
#define kColorCoral (0xff7f50)
#define kColorCyan (0xffff)
#define kColorDarkGreen (0x6400)
#define kColorDarkKHAKI (0xbdb76b)
#define kColorDarkOliveGreen (0x556b2f)
#define kColorDarkOrange (0xff8c00)
#define kColorDodgerBlue (0x1e90ff)
#define kColorGoldenrod (0xdaa520)
#define kColorDarkGray (0x808080)
#define kColorGray (0xb0b0b0)
#define kColorLightGray (0xe0e0e0)
#define kColorGreen (0xff00)
#define kColorKhaki (0xf0e68c)
#define kColorLemonYellow (0xfde910)
#define kColorGold (0xffd700)
#define kColorMetallicGold (0xd4af37)
#define kColorMagenta (0xff00ff)
#define kColorMaroon (0x800000)
#define kColorNavyBlue (0x80)
#define kColorOrange (0xffa500)
#define kColorPink (0xf77fbe)
#define kColorPurple (0xa020f0)
#define kColorRed (0xff0000)
#define kColorBurgundy (0x800020)
#define kColorRoyalBlue (0x4169e1)
#define kColorSalmon (0xfa8072)
#define kColorTurquoise (0x40e0d0)
#define kColorViolet (0xee82ee)
#define kColorWhite (0xffffff)
#define kColorYellow (0xffff00)

#define kColorGray6 (0x606060)
#define kColorGray8 (0x808080)
#define kColorGray10 (0xa0a0a0)
#define kColorGray12 (0xc0c0c0)
#define kColorGray14 (0xe0e0e0)

#endif
