/*=========================================================================
 * This file is part of FrugalWidgets, which is a user interface library 
 * that supports multiple platforms and animation.
 *
 * (C) 2004-2019 Zack T Smith, 1@zsmith.co. All rights reserved.
 *=======================================================================*/

#ifndef _FW_KEYCODES
#define _FW_KEYCODES

enum {
	kKeycode_0='0',
	kKeycode_1,
	kKeycode_2,
	kKeycode_3,
	kKeycode_4,
	kKeycode_5,
	kKeycode_6,
	kKeycode_7,
	kKeycode_8,
	kKeycode_9,
	kKeycode_BackSpace = 8,
	kKeycode_Escape=27,
	kKeycode_Return=10,
	kKeycode_A='A',
	kKeycode_B,
	kKeycode_C,
	kKeycode_D,
	kKeycode_E,
	kKeycode_F,
	kKeycode_G,
	kKeycode_H,
	kKeycode_I,
	kKeycode_J,
	kKeycode_K,
	kKeycode_L,
	kKeycode_M,
	kKeycode_N,
	kKeycode_O,
	kKeycode_P,
	kKeycode_Q,
	kKeycode_R,
	kKeycode_S,
	kKeycode_T,
	kKeycode_U,
	kKeycode_V,
	kKeycode_W,
	kKeycode_X,
	kKeycode_Y,
	kKeycode_Z,
	kKeycode_a = 'a',
	kKeycode_b,
	kKeycode_c,
	kKeycode_d,
	kKeycode_e,
	kKeycode_f,
	kKeycode_g,
	kKeycode_h,
	kKeycode_i,
	kKeycode_j,
	kKeycode_k,
	kKeycode_l,
	kKeycode_m,
	kKeycode_n,
	kKeycode_o,
	kKeycode_p,
	kKeycode_q,
	kKeycode_r,
	kKeycode_s,
	kKeycode_t,
	kKeycode_u,
	kKeycode_v,
	kKeycode_w,
	kKeycode_x,
	kKeycode_y,
	kKeycode_z,
	kKeycode_slash = '/',
	kKeycode_period = '.',
	kKeycode_minus = '-',
	kKeycode_grave = '`',
	kKeycode_equal = '=',
	kKeycode_backslash = '\\',
	kKeycode_comma = ',',
	kKeycode_Tilde = '~',
	kKeycode_Delete = 127,

	kKeycode_Shift_L = 2000,
	kKeycode_Shift_R,
	kKeycode_Control_L,
	kKeycode_Control_R,
        kKeycode_Down,
        kKeycode_Left,
        kKeycode_Right,
        kKeycode_Tab,
        kKeycode_Up,
	kKeycode_Alt_L,
	kKeycode_Meta_L,
	kKeycode_Alt_R,
	kKeycode_Meta_R,
	kKeycode_PageUp,
	kKeycode_PageDown,
	kKeycode_Home,
	kKeycode_Insert,
	kKeycode_End,
	kKeycode_F1,
	kKeycode_F2,
	kKeycode_F3,
	kKeycode_F4,
	kKeycode_F5,
	kKeycode_F6,
	kKeycode_F7,
	kKeycode_F8,
	kKeycode_F9,
	kKeycode_F10,
	kKeycode_F11,
	kKeycode_F12,
	kKeycode_ScrollLock,
	kKeycode_NumLock,
	kKeycode_CapsLock,
	kKeycode_PrintScreen,
};

#endif
