/****************************************************************************
*    Copyright © 2014-2025 The Tumultuous Unicorn Of Darkness
*
*    This program is free software: you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation, either version 3 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*
****************************************************************************/

/*
* PROJECT CPU-X
* FILE core/dmidecode/libdmidecode.h
*/

#ifndef _LIBDMIDECODE_H_
#define _LIBDMIDECODE_H_

#ifdef __cplusplus
extern "C" {
#endif


int dmidecode(int argc, char *const argv[], int quiet, void *cpux_pdata);


#ifdef __cplusplus
} /* extern "C" */
#endif

#endif
